import dataclasses
from typing import *

import hippolyzer.lib.base.serialization as se
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.helpers import get_resource_filename
from hippolyzer.lib.base.templates import AssetType


@dataclasses.dataclass
class VFSBlock:
    location: int = se.dataclass_field(se.U32)
    length: int = se.dataclass_field(se.U32)
    access_time: int = se.dataclass_field(se.U32)
    file_id: UUID = se.dataclass_field(se.UUID)
    asset_type: AssetType = se.dataclass_field(se.IntEnum(AssetType, se.U16))
    size: int = se.dataclass_field(se.U32)


class VFS:
    def __init__(self, index_path):
        self._data_fh = None
        self.blocks: List[VFSBlock] = []
        self._uuid_lookup: Dict[UUID, VFSBlock] = {}

        assert "index.db2" in index_path
        self._index_path = index_path
        self.reload()

    def reload(self):
        self.blocks.clear()
        self._uuid_lookup.clear()
        if self._data_fh:
            self._data_fh.close()
            self._data_fh = None
        self._data_fh = open(self._index_path.replace("index.db2", "data.db2"), "rb")

        with open(self._index_path, "rb") as index_fh:
            reader = se.FHReader("<", index_fh)
            while reader:
                block: VFSBlock = reader.read(se.Dataclass(VFSBlock))
                if not block.size:
                    continue
                self.blocks.append(block)
                self._uuid_lookup[block.file_id] = block

    def __iter__(self) -> Iterator[VFSBlock]:
        return iter(self.blocks)

    def __getitem__(self, item: UUID) -> VFSBlock:
        return self._uuid_lookup[item]

    def __contains__(self, item: UUID):
        return item in self._uuid_lookup

    def __del__(self):
        if self._data_fh:
            self._data_fh.close()
            self._data_fh = None

    def read_block(self, block: VFSBlock) -> bytes:
        self._data_fh.seek(block.location)
        return self._data_fh.read(block.size)


_static_path = get_resource_filename("lib/base/data/static_index.db2")
STATIC_VFS = VFS(_static_path)



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

"""
Exceptions for the hippolyzer library
"""


class Error(Exception):
    """base exception for all hippolyzer related exceptions"""


class AgentError(Error):
    """ general Agent exception """


class LoginError(Error):
    """ raised if there is an error in login """

    def __init__(self, error):
        self.error = error

    def __str__(self):
        return "Agent login error: %s" % self.error


class ParseStartLocError(LoginError):
    """ raised when the is an error in parsing the user specified start location """

    def __init__(self, error):
        self.error = error

    def __str__(self):
        return "Invalid start location specified, using defaults in settings"


# Network errors

class NetworkError(Error):
    """general network exception"""


class HTTPError(NetworkError):
    """an HTTP error"""

    def __init__(self, code, msg, fp, details=""):
        """initialize this exception"""
        self.code = code
        self.msg = msg
        self.details = details
        self.fp = fp

    def __str__(self):
        """return a string representation"""
        return "%s %s" % (self.code, self.msg)


class ResourceNotFound(NetworkError):
    """raised if a resource couldn't be found

    the URL to that resource is stored inside a ``url`` attribute.

    """

    def __init__(self, url=''):
        self.url = url

    def __str__(self):
        return self.url


class ResourceError(NetworkError):
    """raised if any other error occurred (usually a 500)

    contains ``url`` to the resource and ``code`` and ``message`` and ``body``

    """

    def __init__(self, url='', code='', message='', body='', method='GET'):
        self.url = url
        self.code = code
        self.message = message
        self.body = body
        self.method = method

    def __str__(self):
        """return a printable version"""
        return "Error using '%s' on resource '%s': %s (%s)" % (self.method, self.url, self.message, self.code)


# Serialization errors

class SerializationError(Error):
    """serialization related exceptions"""


# Deserialization errors

class DeserializationError(Error):
    """deserialization related exception"""


class DeserializerNotFound(DeserializationError):
    """raised if a deserializer for a certain content type couldn't be found

    stores the content type inside a ``content_type`` attribute.
    """

    def __init__(self, content_type=''):
        self.content_type = content_type

    def __str__(self):
        return "deserialization for '%s' not supported" % self.content_type


class CredentialDeserializerNotFound(DeserializationError):
    """raised if a deserializer for a certain content type couldn't be found

    stores the content type inside a ``content_type`` attribute.
    """

    def __init__(self, payload=''):
        self.payload = payload

    def __str__(self):
        return "deserialization for payload '%s' not supported" % self.payload


class DeserializationFailed(DeserializationError):
    """raised if a deserializer couldn't deserialize a payload

    stores the payload inside a ``payload`` attribute and the error message
    inside a ``reason`` attribute.
    """

    def __init__(self, payload='', reason=''):
        self.payload = payload
        self.reason = reason

    def __str__(self):
        return "deserialization failed for '%s', reason: '%s'" % (self.payload, self.reason)


# Message System related errors

class MessageSystemError(Error):
    """message system related exception"""


class MessageTemplateNotFound(MessageSystemError):
    """ message template file not found

    stores the context in a ``context`` attribute
    """

    def __init__(self, context='', template=''):
        self.context = context
        self.template = template

    def __str__(self):
        return "No message template found for %s, context: '%s'" % (self.template, self.context)


class MessageTemplateParsingError(MessageSystemError):
    """ message template parsing error

    stores the context in a ``context`` attribute
    """

    def __init__(self, context='', template=''):
        self.context = context
        self.template = template

    def __str__(self):
        return "Error parsing message template, context: '%s'" % self.context


class CircuitNotFound(MessageSystemError):
    """ circuit to host could not be found

    stores the host missing a circuit in a ``host`` attribute
    """

    def __init__(self, host='', reason=''):
        self.host = host
        self.reason = reason

    def __str__(self):
        return "No circuit to '%s' found, reason: '%s'" % (self.host, self.reason)


class MessageBuildingError(MessageSystemError):
    """ problem serializing packet data

    stores the label and reason in ``label`` and ``reason`` attributes
    """

    def __init__(self, label='', reason=''):
        self.label = label
        self.reason = reason

    def __str__(self):
        return "Error serializing '%s' due to reason: '%s'" % (self.label, self.reason)


class MessageSerializationError(MessageSystemError):
    """ problem serializing packet data

    stores the label and reason in ``label`` and ``reason`` attributes
    """

    def __init__(self, label='', reason=''):
        self.label = label
        self.reason = reason

    def __str__(self):
        return "Error serializing '%s' due to reason: '%s'" % (self.label, self.reason)


class MessageDeserializationError(MessageSystemError):
    """ problem deserializing packet data

    stores the label and reason in ``label`` and ``reason`` attributes
    """

    def __init__(self, label='', reason=''):
        self.label = label
        self.reason = reason

    def __str__(self):
        return "Error deserializing '%s' due to reason: '%s'" % (self.label, self.reason)


class DataPackingError(MessageSystemError):
    """ problem deserializing packet data

    stores the label and reason in ``label`` and ``reason`` attributes
    """

    def __init__(self, data='', reason=''):
        self.data = data
        self.reason = reason

    def __str__(self):
        return "Error serializing '%s' due to reason: '%s'" % (self.data, self.reason)

    ##########################


# high level exceptions
##########################

# Agent Domain related errors

class AgentDomainError(Error):
    """base exception for all errors which can occur on an agent domain"""


class UserNotFound(AgentDomainError):
    """user couldn't be found

    This exception stores the credentials used inside a ``credentials`` attribute
    """

    def __init__(self, credentials=None):
        """initialize this exception"""
        self.credentials = credentials


class UserNotAuthorized(AgentDomainError):
    """an error raised in case a user couldn't be authorized

    stores the credentials used inside a ``credentials`` attribute

    """

    def __init__(self, credentials=None):
        """initialize this exception"""
        self.credentials = credentials


class UserRezFailed(AgentDomainError):
    """an error raised in case a user couldn't rez on a sim

    stores the region used inside a ``region`` attribute


    """

    def __init__(self, region=None):
        """initialize this exception"""
        self.region = region


# Region Domain related errors

class RegionDomainError(Error):
    """base exception for all errors which can occur on an region domain"""


class RegionSeedCapNotAvailable(RegionDomainError):
    """an error raised in case a user couldn't be authorized

    stores the cap used inside a ``cap`` attribute

    """

    def __init__(self, reason=None):
        """initialize this exception"""

        self.reason = reason

    def __str__(self):
        return "Region seed capability not found when %s'" % self.reason


class RegionCapNotAvailable(RegionDomainError):
    """an error raised in case a user couldn't be authorized

    stores the cap used inside a ``cap`` attribute

    """

    def __init__(self, cap=None):
        """initialize this exception"""

        self.cap = cap


class RegionMessageError(RegionDomainError):
    """ an error raised when a region does not have a connection
    over which it can send UDP messages

    accepts a region object as an attribute

    """

    def __init__(self, region=None):
        """initialize this exception"""

        self.region = region

    def __str__(self):
        return "Agent has no udp connection enabled for: %s'" % self.region


# General errors

class DataParsingError(Error):
    """ an error raised data parsing fails"""

    def __init__(self, error=None):
        """initialize this exception"""

        self.error = error

    def __str__(self):
        return "Data parsing error: %s'" % self.error


class Deprecated(Error):
    """ an error raised when method is not implemented """

    def __init__(self, thing=None):
        """ initialize this exception """

        self.thing = thing

    def __str__(self):
        return "This thing is deprecated: %s" % self.thing



"""
WIP LLMesh -> glTF converter, for testing eventual glTF -> LLMesh conversion logic.
"""
# TODO:
#  * Simple tests
#  * Round-tripping skinning data from Blender-compatible glTF back to LLMesh (maybe through rig retargeting?)
#  * Panda3D-glTF viewer for LLMesh? The glTFs seem to work fine in Panda3D-glTF's `gltf-viewer`.
#  * Check if skew and projection components of transform matrices are ignored in practice as the spec requires.
#    I suppose this would render some real assets impossible to represent with glTF.

import dataclasses
import math
import pprint
import sys
import uuid
from pathlib import Path
from typing import *

import gltflib
import numpy as np
import transformations

from hippolyzer.lib.base.datatypes import Vector3
from hippolyzer.lib.base.mesh import (
    LLMeshSerializer, MeshAsset, positions_from_domain, SkinSegmentDict, VertexWeight, llsd_to_mat4
)
from hippolyzer.lib.base.mesh_skeleton import AVATAR_SKELETON
from hippolyzer.lib.base.serialization import BufferReader


class IdentityList(list):
    """
    List, but does index() by object identity, not equality

    GLTF references objects by their index within some list, but we prefer to pass around
    actual object references internally. If we don't do this, then when we try and get
    a GLTF reference to a given object via `.index()` then we could end up actually getting
    a reference to some other object that just happens to be equal. This was causing issues
    with all primitives ending up with the same material, due to the default material's value
    being the same across all primitives.
    """
    def index(self, value, start: Optional[int] = None, stop: Optional[int] = None) -> int:
        view = self[start:stop]
        for i, x in enumerate(view):
            if x is value:
                if start:
                    return i + start
                return i
        raise ValueError(value)


def sl_to_gltf_coords(coords):
    """
    SL (X, Y, Z) -> GL (X, Z, Y), as GLTF commandeth

    Note that this will only work when reordering axes, flipping an axis is more complicated.
    """
    return coords[0], coords[2], coords[1], *coords[3:]


def sl_to_gltf_uv(uv):
    """Flip the V coordinate of a UV to match glTF convention"""
    return [uv[0], -uv[1]]


def sl_mat4_to_gltf(mat: np.ndarray) -> List[float]:
    """
    Convert an SL Mat4 to the glTF coordinate system

    This should only be done immediately before storing the matrix in a glTF structure!
    """
    # TODO: This is probably not correct. We definitely need to flip Z but there's
    #  probably a better way to do it.
    decomp = [sl_to_gltf_coords(x) for x in transformations.decompose_matrix(mat)]
    trans = decomp[3]
    decomp[3] = (trans[0], trans[1], -trans[2])
    return list(transformations.compose_matrix(*decomp).flatten(order='F'))


# Mat3 to convert points from SL coordinate space to GLTF coordinate space
POINT_TO_GLTF_MAT = transformations.compose_matrix(angles=(-(math.pi / 2), 0, 0))[:3, :3]


def sl_vec3_array_to_gltf(vec_list: np.ndarray) -> np.ndarray:
    new_array = []
    for x in vec_list:
        new_array.append(POINT_TO_GLTF_MAT.dot(x))
    return np.array(new_array)


def sl_weights_to_gltf(sl_weights: List[List[VertexWeight]]) -> Tuple[np.ndarray, np.ndarray]:
    """Convert SL Weights to separate JOINTS_0 and WEIGHTS_0 vec4 arrays"""
    joints = np.zeros((len(sl_weights), 4), dtype=np.uint8)
    weights = np.zeros((len(sl_weights), 4), dtype=np.float32)

    for i, vert_weights in enumerate(sl_weights):
        # We need to re-normalize these since the quantization can mess them up
        collected_weights = []
        for j, vert_weight in enumerate(vert_weights):
            joints[i, j] = vert_weight.joint_idx
            collected_weights.append(vert_weight.weight)
        weight_sum = sum(collected_weights)
        if weight_sum:
            for j, weight in enumerate(collected_weights):
                weights[i, j] = weight / weight_sum

    return joints, weights


def normalize_vec3(a):
    norm = np.linalg.norm(a)
    if norm == 0:
        return a
    return a / norm


def apply_bind_shape_matrix(bind_shape_matrix: np.ndarray, verts: np.ndarray, norms: np.ndarray) \
        -> Tuple[np.ndarray, np.ndarray]:
    """
    Apply the bind shape matrix to the mesh data

    glTF expects all verts and normals to be in armature-local space so that mesh data can be shared
    between differently-oriented armatures. Or something.
    # https://github.com/KhronosGroup/glTF-Blender-IO/issues/566#issuecomment-523119339

    glTF also doesn't have a concept of a "bind shape matrix" like Collada does
    per its skinning docs, so we have to mix it into the mesh data manually.
    See https://github.com/KhronosGroup/glTF-Tutorials/blob/master/gltfTutorial/gltfTutorial_020_Skins.md
    """
    scale, _, angles, translation, _ = transformations.decompose_matrix(bind_shape_matrix)
    scale_mat = transformations.compose_matrix(scale=scale)[:3, :3]
    rot_mat = transformations.euler_matrix(*angles)[:3, :3]
    rot_scale_mat = scale_mat @ np.linalg.inv(rot_mat)

    # Apply the SRT transform to each vert
    verts = (verts @ rot_scale_mat) + translation

    # Our scale is unlikely to be uniform, so we have to fix up our normals as well.
    # https://paroj.github.io/gltut/Illumination/Tut09%20Normal%20Transformation.html
    inv_transpose_mat = np.transpose(np.linalg.inv(bind_shape_matrix)[:3, :3])
    new_norms = [normalize_vec3(inv_transpose_mat @ norm) for norm in norms]

    return verts, np.array(new_norms)


@dataclasses.dataclass
class JointContext:
    node: gltflib.Node
    # Original matrix for the bone, may have custom translation, but otherwise the same.
    orig_matrix: np.ndarray
    # xform that must be applied to inverse bind matrices to account for the changed bone
    fixup_matrix: np.ndarray


JOINT_CONTEXT_DICT = Dict[str, JointContext]


class GLTFBuilder:
    def __init__(self, blender_compatibility=False):
        self.scene = gltflib.Scene(nodes=IdentityList())
        self.model = gltflib.GLTFModel(
            asset=gltflib.Asset(version="2.0"),
            accessors=IdentityList(),
            nodes=IdentityList(),
            materials=IdentityList(),
            buffers=IdentityList(),
            bufferViews=IdentityList(),
            meshes=IdentityList(),
            skins=IdentityList(),
            scenes=IdentityList((self.scene,)),
            extensionsUsed=["KHR_materials_specular"],
            scene=0,
        )
        self.gltf = gltflib.GLTF(
            model=self.model,
            resources=IdentityList(),
        )
        self.blender_compatibility = blender_compatibility

    def add_nodes_from_llmesh(self, mesh: MeshAsset, name: str, mesh_transform: Optional[np.ndarray] = None):
        """Build a glTF version of a mesh asset, appending it and its armature to the scene root"""
        # TODO: mesh data instancing?
        #  consider https://github.com/KhronosGroup/glTF-Blender-IO/issues/1634.
        if mesh_transform is None:
            mesh_transform = np.identity(4)

        skin_seg: Optional[SkinSegmentDict] = mesh.segments.get('skin')
        skin = None
        if skin_seg:
            mesh_transform = llsd_to_mat4(skin_seg['bind_shape_matrix'])
            joint_ctxs = self.add_joints(skin_seg)

            # Give our armature a root node and parent the pelvis to it
            armature_node = self.add_node("Armature")
            self.scene.nodes.append(self.model.nodes.index(armature_node))
            armature_node.children.append(self.model.nodes.index(joint_ctxs['mPelvis'].node))
            skin = self.add_skin("Armature", joint_ctxs, skin_seg)
            skin.skeleton = self.model.nodes.index(armature_node)

        primitives = []
        # Just the high LOD for now
        for submesh in mesh.segments['high_lod']:
            verts = np.array(positions_from_domain(submesh['Position'], submesh['PositionDomain']))
            norms = np.array(submesh['Normal'])
            tris = np.array(submesh['TriangleList'])
            joints = np.array([])
            weights = np.array([])
            range_uv = np.array([])
            if "TexCoord0" in submesh:
                range_uv = np.array(positions_from_domain(submesh['TexCoord0'], submesh['TexCoord0Domain']))
            if 'Weights' in submesh:
                joints, weights = sl_weights_to_gltf(submesh['Weights'])

            if skin:
                # Convert verts and norms to armature-local space
                verts, norms = apply_bind_shape_matrix(mesh_transform, verts, norms)

            primitives.append(self.add_primitive(
                tris=tris,
                positions=verts,
                normals=norms,
                uvs=range_uv,
                joints=joints,
                weights=weights,
            ))

        mesh_node = self.add_node(
            name,
            self.add_mesh(name, primitives),
            transform=mesh_transform,
        )
        if skin:
            # Node translation isn't relevant, we're going to use the bind matrices
            # If you pull this into Blender you may want to untick "Guess Original Bind Pose",
            # it guesses that based on the inverse bind matrices which may have Maya poisoning.
            # TODO: Maybe we could automatically undo that by comparing expected bone scale and rot
            #  to scale and rot in the inverse bind matrices, and applying fixups to the
            #  bind shape matrix and inverse bind matrices?
            mesh_node.matrix = None
            mesh_node.skin = self.model.skins.index(skin)

        self.scene.nodes.append(self.model.nodes.index(mesh_node))

    def add_node(
            self,
            name: str,
            mesh: Optional[gltflib.Mesh] = None,
            transform: Optional[np.ndarray] = None,
    ) -> gltflib.Node:
        node = gltflib.Node(
            name=name,
            mesh=self.model.meshes.index(mesh) if mesh else None,
            matrix=sl_mat4_to_gltf(transform) if transform is not None else None,
            children=[],
        )
        self.model.nodes.append(node)
        return node

    def add_mesh(
            self,
            name: str,
            primitives: List[gltflib.Primitive],
    ) -> gltflib.Mesh:
        for i, prim in enumerate(primitives):
            # Give the materials a name relating to what "face" they belong to
            self.model.materials[prim.material].name = f"{name}.{i:03}"
        mesh = gltflib.Mesh(name=name, primitives=primitives)
        self.model.meshes.append(mesh)
        return mesh

    def add_primitive(
            self,
            tris: np.ndarray,
            positions: np.ndarray,
            normals: np.ndarray,
            uvs: np.ndarray,
            weights: np.ndarray,
            joints: np.ndarray,
    ) -> gltflib.Primitive:
        # Make a Material for the primitive. Materials pretty much _are_ the primitives in
        # LLMesh, so just make them both in one go. We need a unique material for each primitive.
        material = gltflib.Material(
            pbrMetallicRoughness=gltflib.PBRMetallicRoughness(
                baseColorFactor=[1.0, 1.0, 1.0, 1.0],
                metallicFactor=0.0,
                roughnessFactor=0.0,
            ),
            extensions={
                "KHR_materials_specular": {
                    "specularFactor": 0.0,
                    "specularColorFactor": [0, 0, 0]
                },
            }
        )
        self.model.materials.append(material)

        attributes = gltflib.Attributes(
            POSITION=self.maybe_add_vec_array(sl_vec3_array_to_gltf(positions), gltflib.AccessorType.VEC3),
            NORMAL=self.maybe_add_vec_array(sl_vec3_array_to_gltf(normals), gltflib.AccessorType.VEC3),
            TEXCOORD_0=self.maybe_add_vec_array(np.array([sl_to_gltf_uv(uv) for uv in uvs]), gltflib.AccessorType.VEC2),
            JOINTS_0=self.maybe_add_vec_array(joints, gltflib.AccessorType.VEC4, gltflib.ComponentType.UNSIGNED_BYTE),
            WEIGHTS_0=self.maybe_add_vec_array(weights, gltflib.AccessorType.VEC4),
        )

        return gltflib.Primitive(
            attributes=attributes,
            indices=self.model.accessors.index(self.add_scalars(tris)),
            material=self.model.materials.index(material),
            mode=gltflib.PrimitiveMode.TRIANGLES,
        )

    def add_scalars(self, scalars: np.ndarray) -> gltflib.Accessor:
        """
        Add a potentially multidimensional array of scalars, returning the accessor

        Generally only used for triangle indices
        """
        scalar_bytes = scalars.astype(np.uint32).flatten().tobytes()
        buffer_view = self.add_buffer_view(scalar_bytes, None)
        accessor = gltflib.Accessor(
            bufferView=self.model.bufferViews.index(buffer_view),
            componentType=gltflib.ComponentType.UNSIGNED_INT,
            count=scalars.size,  # use the flattened size!
            type=gltflib.AccessorType.SCALAR.value,  # type: ignore
            min=[int(scalars.min())],  # type: ignore
            max=[int(scalars.max())],  # type: ignore
        )
        self.model.accessors.append(accessor)
        return accessor

    def maybe_add_vec_array(
            self,
            vecs: np.ndarray,
            vec_type: gltflib.AccessorType,
            component_type: gltflib.ComponentType = gltflib.ComponentType.FLOAT,
    ) -> Optional[int]:
        if not vecs.size:
            return None
        accessor = self.add_vec_array(vecs, vec_type, component_type)
        return self.model.accessors.index(accessor)

    def add_vec_array(
            self,
            vecs: np.ndarray,
            vec_type: gltflib.AccessorType,
            component_type: gltflib.ComponentType = gltflib.ComponentType.FLOAT
    ) -> gltflib.Accessor:
        """
        Add a two-dimensional array of vecs (positions, normals, weights, UVs) returning the accessor

        Vec type may be a vec2, vec3, or a vec4.
        """
        # Pretty much all of these are float32 except the ones that aren't
        dtype = np.float32
        if component_type == gltflib.ComponentType.UNSIGNED_BYTE:
            dtype = np.uint8
        vec_data = vecs.astype(dtype).tobytes()
        buffer_view = self.add_buffer_view(vec_data, target=None)
        accessor = gltflib.Accessor(
            bufferView=self.model.bufferViews.index(buffer_view),
            componentType=component_type,
            count=len(vecs),
            type=vec_type.value,  # type: ignore
            min=vecs.min(axis=0).tolist(),  # type: ignore
            max=vecs.max(axis=0).tolist(),  # type: ignore
        )
        self.model.accessors.append(accessor)
        return accessor

    def add_buffer_view(self, data: bytes, target: Optional[gltflib.BufferTarget]) -> gltflib.BufferView:
        """Create a buffer view and associated buffer and resource for a blob of data"""
        resource = gltflib.FileResource(filename=f"res-{uuid.uuid4()}.bin", data=data)
        self.gltf.resources.append(resource)

        buffer = gltflib.Buffer(uri=resource.filename, byteLength=len(resource.data))
        self.model.buffers.append(buffer)

        buffer_view = gltflib.BufferView(
            buffer=self.model.buffers.index(buffer),
            byteLength=buffer.byteLength,
            byteOffset=0,
            target=target
        )
        self.model.bufferViews.append(buffer_view)
        return buffer_view

    def add_joints(self, skin: SkinSegmentDict) -> JOINT_CONTEXT_DICT:
        # There may be some joints not present in the mesh that we need to add to reach the mPelvis root
        required_joints = set()
        for joint_name in skin['joint_names']:
            joint_node = AVATAR_SKELETON[joint_name]
            required_joints.add(joint_node)
            required_joints.update(joint_node.ancestors)

        # If this is present, it may override the joint positions from the skeleton definition
        if 'alt_inverse_bind_matrix' in skin:
            joint_overrides = dict(zip(skin['joint_names'], skin['alt_inverse_bind_matrix']))
        else:
            joint_overrides = {}

        built_joints: JOINT_CONTEXT_DICT = {}
        for joint in required_joints:
            joint_matrix = joint.matrix

            # Do we have a joint position override that would affect joint_matrix?
            override = joint_overrides.get(joint.name)
            if override:
                decomp = list(transformations.decompose_matrix(joint_matrix))
                # We specifically only want the translation from the override!
                translation = transformations.translation_from_matrix(llsd_to_mat4(override))
                # Only do it if the difference is over 0.1mm though
                if Vector3.dist(Vector3(*translation), joint.translation) > 0.0001:
                    decomp[3] = translation
                    joint_matrix = transformations.compose_matrix(*decomp)

            # Do we need to mess with the bone's matrices to make Blender cooperate?
            orig_matrix = joint_matrix
            fixup_matrix = np.identity(4)
            if self.blender_compatibility:
                joint_matrix, fixup_matrix = self._fix_blender_joint(joint_matrix)

            # TODO: populate "extras" here with the metadata the Blender collada stuff uses to store
            #  "bind_mat" and "rest_mat" so we can go back to our original matrices when exporting
            #  from blender to .dae!
            gltf_joint = self.add_node(joint.name, transform=joint_matrix)

            # Store the node along with any fixups we may need to apply to the bind matrices later
            built_joints[joint.name] = JointContext(gltf_joint, orig_matrix, fixup_matrix)

        # Add each joint to the child list of their respective parent
        for joint_name, joint_ctx in built_joints.items():
            if parent := AVATAR_SKELETON[joint_name].parent:
                built_joints[parent().name].node.children.append(self.model.nodes.index(joint_ctx.node))
        return built_joints

    def _fix_blender_joint(self, joint_matrix: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Split a joint matrix into a joint matrix and fixup matrix

        If we don't account for weird scaling on the collision volumes, then
        Blender freaks out. This is an issue in blender where it doesn't
        apply the inverse bind matrices relative to the scale and rotation of
        the bones themselves, as it should per the glTF spec. Blender's glTF loader
        tries to recover from this by applying certain transforms as a pose, but
        the damage has been done by that point. Nobody else runs really runs into
        this because they have the good sense to not use some nightmare abomination
        rig with scaling and rotation on the skeleton like SL does.

        Blender will _only_ correctly handle the translation component of the joint,
        any other transforms need to be mixed into the inverse bind matrices themselves.
        There's no internal concept of bone scale or rot in Blender right now.

        Should investigate an Avastar-style approach of optionally retargeting
        to a Blender-compatible rig with translation-only bones, and modify
        the bind matrices to accommodate. The glTF importer supports metadata through
        the "extras" fields, so we can potentially abuse the "bind_mat" metadata field
        that Blender already uses for the "Keep Bind Info" Collada import / export hack.

        For context:
        * https://github.com/KhronosGroup/glTF-Blender-IO/issues/1305
        * https://developer.blender.org/T38660 (these are Collada, but still relevant)
        * https://developer.blender.org/T29246
        * https://developer.blender.org/T50412
        * https://developer.blender.org/T53620 (FBX but still relevant)
        """
        scale, shear, angles, translate, projection = transformations.decompose_matrix(joint_matrix)
        joint_matrix = transformations.compose_matrix(translate=translate)
        fixup_matrix = transformations.compose_matrix(scale=scale, angles=angles)
        return joint_matrix, fixup_matrix

    def add_skin(self, name: str, joint_nodes: JOINT_CONTEXT_DICT, skin_seg: SkinSegmentDict) -> gltflib.Skin:
        joints_arr = []
        for joint_name in skin_seg['joint_names']:
            joint_ctx = joint_nodes[joint_name]
            joints_arr.append(self.model.nodes.index(joint_ctx.node))

        inv_binds = []
        for joint_name, inv_bind in zip(skin_seg['joint_names'], skin_seg['inverse_bind_matrix']):
            joint_ctx = joint_nodes[joint_name]
            inv_bind = joint_ctx.fixup_matrix @ llsd_to_mat4(inv_bind)
            inv_binds.append(sl_mat4_to_gltf(inv_bind))
        inv_binds_data = np.array(inv_binds, dtype=np.float32).tobytes()
        buffer_view = self.add_buffer_view(inv_binds_data, target=None)
        accessor = gltflib.Accessor(
            bufferView=self.model.bufferViews.index(buffer_view),
            componentType=gltflib.ComponentType.FLOAT,
            count=len(inv_binds),
            type=gltflib.AccessorType.MAT4.value,  # type: ignore
        )
        self.model.accessors.append(accessor)
        accessor_idx = self.model.accessors.index(accessor)

        skin = gltflib.Skin(name=name, joints=joints_arr, inverseBindMatrices=accessor_idx)
        self.model.skins.append(skin)
        return skin

    def finalize(self):
        """Clean up the mesh to pass the glTF smell test, should be done last"""
        def _nullify_empty_lists(dc):
            for field in dataclasses.fields(dc):
                # Empty lists should be replaced with None
                if getattr(dc, field.name) == []:
                    setattr(dc, field.name, None)

        for node in self.model.nodes:
            _nullify_empty_lists(node)
        _nullify_empty_lists(self.model)
        return self.gltf


def main():
    # Take an llmesh file as an argument and spit out basename-converted.gltf
    with open(sys.argv[1], "rb") as f:
        reader = BufferReader("<", f.read())

    filename = Path(sys.argv[1]).stem
    mesh: MeshAsset = reader.read(LLMeshSerializer(parse_segment_contents=True))

    builder = GLTFBuilder(blender_compatibility=True)
    builder.add_nodes_from_llmesh(mesh, filename)
    gltf = builder.finalize()

    pprint.pprint(gltf.model)
    gltf.export_glb(sys.argv[1].rsplit(".", 1)[0] + "-converted.gltf")


if __name__ == "__main__":
    main()



import asyncio
import logging
import weakref
from typing import Optional, Tuple

from hippolyzer.lib.base.message.message_dot_xml import MessageDotXML
from hippolyzer.lib.base.message.udpdeserializer import UDPMessageDeserializer
from hippolyzer.lib.base.message.udpserializer import UDPMessageSerializer
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.base.network.transport import UDPPacket
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager
from hippolyzer.lib.proxy.socks_proxy import SOCKS5Server, UDPProxyProtocol


LOG = logging.getLogger(__name__)


class SLSOCKS5Server(SOCKS5Server):
    def __init__(self, session_manager=None):
        super().__init__()
        self.session_manager = weakref.proxy(session_manager)

    def _udp_protocol_creator(self, source_addr):
        return lambda: InterceptingLLUDPProxyProtocol(source_addr, self.session_manager)


class InterceptingLLUDPProxyProtocol(UDPProxyProtocol):
    def __init__(self, source_addr: Tuple[str, int], session_manager: SessionManager):
        super().__init__(source_addr)
        self.session_manager: SessionManager = session_manager
        self.serializer = UDPMessageSerializer()
        self.deserializer = UDPMessageDeserializer(
            settings=self.session_manager.settings,
        )
        self.message_xml = MessageDotXML()
        self.session: Optional[Session] = None
        loop = asyncio.get_event_loop_policy().get_event_loop()
        self.resend_task = loop.create_task(self.attempt_resends())

    async def attempt_resends(self):
        while True:
            await asyncio.sleep(0.1)
            if self.session is None:
                continue
            for region in self.session.regions:
                if not region.circuit or not region.circuit.is_alive:
                    continue
                region.circuit.resend_unacked()

    def _ensure_message_allowed(self, msg: Message):
        if not self.message_xml.validate_udp_msg(msg.name):
            LOG.warning(
                f"Received {msg.name!r} over UDP, when it should come over the event queue. Discarding."
            )
            raise PermissionError(f"UDPBanned message {msg.name}")

    def handle_proxied_packet(self, packet: UDPPacket):
        region: Optional[ProxiedRegion] = None
        # Try to do an initial region lookup so we have it for handle_proxied_packet()
        if self.session:
            region = self.session.region_by_circuit_addr(packet.far_addr)

        # the proxied packet handler is allowed to mutate `packet.data` before
        # the message gets parsed.
        if AddonManager.handle_proxied_packet(self.session_manager, packet,
                                              self.session, region):
            return

        message = self.deserializer.deserialize(packet.data)
        message.direction = packet.direction
        message.sender = packet.src_addr
        message.meta.update(packet.meta)

        assert message is not None
        # Check for UDP bans on inbound messages
        if packet.incoming:
            self._ensure_message_allowed(message)

        if not self.session:
            # This proxy instance isn't tied to a session yet
            # First message should be "UseCircuitCode" and should let us
            # claim a pending session
            if message.name == "UseCircuitCode" and packet.outgoing:
                session_id = message["CircuitCode"][0]["SessionID"]
                self.session = self.session_manager.claim_session(session_id)
                if not self.session:
                    LOG.error(f"Wasn't able to claim session {session_id!r}! Generally this means that the "
                              f"login HTTP request was not intercepted for some reason. Have you configured "
                              f"the viewer to send all HTTP traffic through this proxy?")
                    return
            else:
                LOG.warning(
                    "Received unexpected message %s on %r before circuit open" % (
                        message.name, packet.far_addr
                    )
                )
                return

        if message.name == "UseCircuitCode" and packet.outgoing:
            # This will create a circuit, replace a circuit, or do nothing if
            # circuit is already alive.
            if not self.session.open_circuit(packet.src_addr, packet.dst_addr, self.transport):
                LOG.warning("Couldn't open circuit to %r, did we have a region???" % (
                    packet.dst_addr,
                ))
                return

        region = self.session.region_by_circuit_addr(packet.far_addr)
        if not region:
            LOG.error("No circuit for %r, dropping packet!" % (packet.far_addr,))
            return

        # Process any ACKs for messages we injected first
        region.circuit.collect_acks(message)

        if message.name == "AgentMovementComplete":
            self.session.main_region = region
            if region.handle is None:
                region.handle = message["Data"]["RegionHandle"]
            LOG.info(f"Setting main region to {region!r}, had circuit addr {packet.far_addr!r}")
            AddonManager.handle_region_changed(self.session, region)
        if message.name == "RegionHandshake":
            region.cache_id = message["RegionInfo"]["CacheID"]
            self.session.objects.track_region_objects(region.handle)
            if self.session_manager.settings.USE_VIEWER_OBJECT_CACHE:
                try:
                    region.objects.load_cache()
                except:
                    LOG.exception("Failed to load region cache, skipping")

        try:
            self.session.message_handler.handle(message)
        except:
            LOG.exception("Failed in session message handler")
        try:
            region.message_handler.handle(message)
        except:
            LOG.exception("Failed in region message handler")

        message_logger = self.session_manager.message_logger

        handled = AddonManager.handle_lludp_message(
            self.session, region, message
        )

        # This message is owned by an async handler, drop it so it doesn't get
        # sent with the normal flow.
        if message.queued:
            region.circuit.drop_message(message)

        # Shouldn't mutate the message past this point, so log it now.
        if message_logger:
            message_logger.log_lludp_message(self.session, region, message)

        if handled:
            return

        if message.name in ("CloseCircuit", "DisableSimulator"):
            region.mark_dead()
        elif message.name == "RegionHandshake":
            region.name = str(message["RegionInfo"][0]["SimName"])

        # Send the message if it wasn't explicitly dropped or sent before
        if not message.finalized:
            region.circuit.send(message)

    def close(self):
        super().close()
        if self.session:
            AddonManager.handle_session_closed(self.session)
            self.session_manager.close_session(self.session)
        self.session = None
        self.resend_task.cancel()



from __future__ import annotations

import asyncio
import copy
import dataclasses
from types import TracebackType
from typing import *

import aiohttp
import multidict

from hippolyzer.lib.base import llsd as llsd_lib


class CapsClientResponse(aiohttp.ClientResponse):
    """
    Not actually instantiated, used for lying to the type system
    since we'll dynamically put this onto a ClientResponse instance
    Will fail isinstance().
    """
    async def read_llsd(self) -> Any:
        raise NotImplementedError()


class _HippoSessionRequestContextManager:
    """
    _SessionRequestContextManager but with a symmetrical API

    aiohttp.request() and aiohttp.ClientSession.request() have different APIs.
    One is sync returning a context manager, one is async returning a coro.
    aiohttp.request() also doesn't accept the arguments that we need for custom
    SSL contexts. To deal with requests that have existing sessions and those without,
    just give them both the same wrapper and don't close the session on context manager
    exit if it wasn't our session.
    """
    __slots__ = ("_coro", "_resp", "_session", "_session_owned")

    def __init__(
        self,
        coro: Coroutine[asyncio.Future[Any], None, aiohttp.ClientResponse],
        session: aiohttp.ClientSession,
        session_owned: bool = True,
    ) -> None:
        self._coro = coro
        self._resp: Optional[aiohttp.ClientResponse] = None
        self._session = session
        self._session_owned = session_owned

    async def __aenter__(self) -> CapsClientResponse:
        try:
            self._resp = await self._coro

            # We don't control creation of the ClientResponse, so tack on
            # a convenience method for reading LLSD.
            async def _read_llsd():
                return llsd_lib.parse_xml(await self._resp.read())
            self._resp.read_llsd = _read_llsd
        except BaseException:
            if self._session_owned:
                await self._session.close()
            raise
        else:
            # intentionally fooling the type system
            return self._resp  # type: ignore

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        assert self._resp is not None
        self._resp.close()
        if self._session_owned:
            await self._session.close()


CAPS_DICT = Union[
    Mapping[str, str],
    multidict.MultiDict[str],
]


class CapsClient:
    def __init__(self, caps: Optional[CAPS_DICT] = None):
        self._caps = caps

    def _request_fixups(self, cap_or_url: str, headers: Dict, proxy: Optional[bool], ssl: Any):
        return cap_or_url, headers, proxy, ssl

    def _get_caps(self) -> Optional[CAPS_DICT]:
        return self._caps

    def request(self, method: str, cap_or_url: str, *, path: str = "", data: Any = None,
                headers: Optional[Dict] = None, session: Optional[aiohttp.ClientSession] = None,
                llsd: Any = dataclasses.MISSING, params: Optional[Dict[str, Any]] = None,
                proxy: Optional[str] = None, skip_auto_headers: Optional[Sequence[str]] = None,
                **kwargs) -> _HippoSessionRequestContextManager:
        if cap_or_url.startswith("http"):
            if path:
                raise ValueError("Specifying both path and a full URL not supported")
        else:
            caps = self._get_caps()
            if caps is None:
                raise RuntimeError(f"Need a caps dict to request a Cap like {cap_or_url}")
            if cap_or_url not in caps:
                raise KeyError(f"{cap_or_url} is not a full URL and not a Cap")
            cap_or_url = caps[cap_or_url]
        if path:
            cap_or_url += path

        if params is not None:
            for pname, pval in params.items():
                if not isinstance(pval, str):
                    params[pname] = str(pval)

        session_owned = False
        # Use an existing session if we have one to take advantage of connection pooling
        # otherwise create one
        if session is None:
            session_owned = True
            session = aiohttp.ClientSession(
                connector=aiohttp.TCPConnector(force_close=True),
                connector_owner=True
            )

        if headers is None:
            headers = {}
        else:
            headers = copy.copy(headers)

        # Use sentinel val so explicit `None` can be passed
        if llsd is not dataclasses.MISSING:
            data = llsd_lib.format_xml(llsd)
        # Sometimes needed even on GETs.
        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/llsd+xml"
        # Always present, usually ignored by the server.
        if "Accept" not in headers:
            headers["Accept"] = "application/llsd+xml"
        # Ask to keep the connection open if we're sharing a session
        if not session_owned:
            headers["Connection"] = "keep-alive"
            headers["Keep-alive"] = "300"

        ssl = kwargs.pop('ssl', None)
        cap_or_url, headers, proxy, ssl = self._request_fixups(cap_or_url, headers, proxy, ssl)

        resp = session._request(method, cap_or_url, data=data, headers=headers,  # noqa: need internal call
                                params=params, ssl=ssl, proxy=proxy,
                                skip_auto_headers=skip_auto_headers or ("User-Agent",), **kwargs)
        return _HippoSessionRequestContextManager(resp, session, session_owned=session_owned)

    def get(self, cap_or_url: str, *, path: str = "", headers: Optional[dict] = None,
            session: Optional[aiohttp.ClientSession] = None, params: Optional[Dict[str, Any]] = None,
            proxy: Optional[str] = None, **kwargs) -> _HippoSessionRequestContextManager:
        return self.request("GET", cap_or_url=cap_or_url, path=path, headers=headers,
                            session=session, params=params, proxy=proxy, **kwargs)

    def post(self, cap_or_url: str, *, path: str = "", data: Any = None,
             headers: Optional[dict] = None, session: Optional[aiohttp.ClientSession] = None,
             llsd: Any = dataclasses.MISSING, params: Optional[Dict[str, Any]] = None,
             proxy: Optional[str] = None, **kwargs) -> _HippoSessionRequestContextManager:
        return self.request("POST", cap_or_url=cap_or_url, path=path, headers=headers, data=data,
                            llsd=llsd, session=session, params=params, proxy=proxy, **kwargs)

    def put(self, cap_or_url: str, *, path: str = "", data: Any = None,
            headers: Optional[dict] = None, session: Optional[aiohttp.ClientSession] = None,
            llsd: Any = dataclasses.MISSING, params: Optional[Dict[str, Any]] = None,
            proxy: Optional[str] = None, **kwargs) -> _HippoSessionRequestContextManager:
        return self.request("PUT", cap_or_url=cap_or_url, path=path, headers=headers, data=data,
                            llsd=llsd, session=session, params=params, proxy=proxy, **kwargs)



"""
Copyright 2008, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

version = '0.13.1'

with open(path.join(here, 'README.md')) as readme_fh:
    readme = readme_fh.read()

setup(
    name='hippolyzer',
    version=version,
    description="Analysis tools for SL-compatible virtual worlds",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: System :: Networking :: Monitoring",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Testing",
    ],
    author='Salad Dais',
    author_email='83434023+SaladDais@users.noreply.github.com',
    url='https://github.com/SaladDais/Hippolyzer/',
    license='LGPLv3',
    packages=find_packages(include=["hippolyzer", "hippolyzer.*"]),
    package_data={
        'hippolyzer': [
            'apps/message_builder.ui',
            'apps/proxy_mainwindow.ui',
            'apps/filter_dialog.ui',
            'apps/addon_dialog.ui',
            'lib/base/message/data/message_template.msg',
            'lib/base/message/data/message.xml',
            'lib/base/network/data/ca-bundle.crt',
            'lib/base/data/static_data.db2',
            'lib/base/data/static_index.db2',
            'lib/base/data/avatar_lad.xml',
            'lib/base/data/male_collada_joints.xml',
            'lib/base/data/avatar_skeleton.xml',
            'lib/base/data/LICENSE-artwork.txt',
        ],
    },
    entry_points={
        'console_scripts': {
            'hippolyzer-gui = hippolyzer.apps.proxy_gui:gui_main',
            'hippolyzer-cli = hippolyzer.apps.proxy:main',
        }
    },
    zip_safe=False,
    python_requires='>=3.8',
    install_requires=[
        'llsd<1.1.0',
        'outleap<1.0',
        'defusedxml',
        'aiohttp<4.0.0',
        # Newer recordclasses break!
        'recordclass>0.15,<0.18.3',
        'lazy-object-proxy',
        'arpeggio',
        # requests breaks with newer idna
        'idna<3,>=2.5',
        # 7.x will be a major change.
        'mitmproxy>=8.0.0,<8.1',
        # For REPLs
        'ptpython<4.0',
        # JP2 codec
        'Glymur<0.9.7',
        'numpy<2.0',
        # These could be in extras_require if you don't want a GUI.
        'pyside6-essentials',
        'qasync',
        # Needed for mesh format conversion tooling
        'pycollada',
        'transformations',
        'gltflib',
    ],
    tests_require=[
        "pytest",
        "aioresponses",
    ],
)



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import typing

from .msgtypes import MsgType, MsgBlockType
from ..datatypes import UUID


class MessageTemplateVariable:
    def __init__(self, name, tp, size):
        self.name = name
        self.type: MsgType = tp
        self.size = size
        self._probably_binary = None
        self._probably_text = None

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name!r}, tp={self.type!r}, size={self.size!r})"

    @property
    def probably_binary(self):
        if self._probably_binary is not None:
            return self._probably_binary

        if self.type not in {MsgType.MVT_FIXED, MsgType.MVT_VARIABLE}:
            self._probably_binary = False
        else:
            self._probably_binary = any(x in self.name for x in
                                        ("Binary", "Data", "Handle", "Color", "Texture", "Params", "NameValue"))
        return self._probably_binary

    @property
    def probably_text(self):
        if self._probably_text is not None:
            return self._probably_text

        if self.type not in {MsgType.MVT_FIXED, MsgType.MVT_VARIABLE}:
            self._probably_text = False
        else:
            self._probably_text = any(x in self.name for x in (
                "Name", "Text", "Title", "Description", "Message", "Label", "Method", "Filename",
            ))
            self._probably_text = self._probably_text and self.name != "NameValue"
        return self._probably_text

    @property
    def default_value(self):
        if self.type.is_int:
            return 0
        elif self.type.is_float:
            return 0.0
        elif self.type == MsgType.MVT_LLUUID:
            return UUID()
        elif self.type == MsgType.MVT_BOOL:
            return False
        elif self.type == MsgType.MVT_VARIABLE:
            if self.probably_binary:
                return b""
            if self.probably_text:
                return ""
            return b""
        elif self.type in (MsgType.MVT_LLVector3, MsgType.MVT_LLVector3d, MsgType.MVT_LLQuaternion):
            return 0.0, 0.0, 0.0
        elif self.type == MsgType.MVT_LLVector4:
            return 0.0, 0.0, 0.0, 0.0
        elif self.type == MsgType.MVT_FIXED:
            return b"\x00" * self.size
        elif self.type == MsgType.MVT_IP_ADDR:
            return "0.0.0.0"
        return None


class MessageTemplateBlock:
    def __init__(self, name):
        self.variables: typing.List[MessageTemplateVariable] = []
        self.variable_map: typing.Dict[str, MessageTemplateVariable] = {}
        self.name = name
        self.block_type: MsgBlockType = MsgBlockType.MBT_SINGLE
        self.number = 0

    def add_variable(self, var):
        self.variable_map[var.name] = var
        self.variables.append(var)

    def get_variable(self, name):
        return self.variable_map[name]


class MessageTemplate(object):
    frequency_strings = {-1: 'fixed', 1: 'high', 2: 'medium', 4: 'low'}  # strings for printout
    deprecation_strings = ["Deprecated", "UDPDeprecated", "UDPBlackListed", "NotDeprecated"]  # using _as_string methods
    encoding_strings = ["Unencoded", "Zerocoded"]  # etc
    trusted_strings = ["Trusted", "NotTrusted"]  # etc LDE 24oct2008

    def __init__(self, name):
        self.blocks: typing.List[MessageTemplateBlock] = []
        self.block_map: typing.Dict[str, MessageTemplateBlock] = {}

        # this is the function or object that will handle this type of message
        self.received_count = 0

        self.name = name
        self.frequency = None
        self.msg_num = 0
        self.msg_freq_num_bytes = None
        self.msg_trust = None
        self.msg_deprecation = None
        self.msg_encoding = None

    def add_block(self, block):
        self.block_map[block.name] = block
        self.blocks.append(block)

    def get_block(self, name):
        return self.block_map[name]

    def get_msg_freq_num_len(self):
        if self.frequency == -1:
            return 4
        return self.frequency

    def get_frequency_as_string(self):
        return MessageTemplate.frequency_strings[self.frequency]

    def get_deprecation_as_string(self):
        return MessageTemplate.deprecation_strings[self.msg_deprecation]



from __future__ import annotations

import dataclasses
import enum
from typing import *

import hippolyzer.lib.base.serialization as se
from hippolyzer.lib.base.datatypes import Vector3, Quaternion
from hippolyzer.lib.base.multidict import OrderedMultiDict


# Forward declarations so the dataclasses can be declared in natural order
JOINT_DATACLASS = se.ForwardSerializable(lambda: se.Dataclass(Joint))
CONSTRAINT_DATACLASS = se.ForwardSerializable(lambda: se.Dataclass(Constraint))
POSKEYFRAME_DATACLASS = se.ForwardSerializable(lambda: se.Dataclass(PosKeyframe))
ROTKEYFRAME_DATACLASS = se.ForwardSerializable(lambda: se.Dataclass(RotKeyframe))


@dataclasses.dataclass
class Animation:
    major_version: int = se.dataclass_field(se.U16, default=1)
    minor_version: int = se.dataclass_field(se.U16, default=0)
    base_priority: int = se.dataclass_field(se.S32)
    duration: float = se.dataclass_field(se.F32)
    emote_name: str = se.dataclass_field(se.CStr())
    loop_in_point: float = se.dataclass_field(se.F32)
    loop_out_point: float = se.dataclass_field(se.F32)
    loop: int = se.dataclass_field(se.S32)
    ease_in_duration: float = se.dataclass_field(se.F32)
    ease_out_duration: float = se.dataclass_field(se.F32)
    hand_pose: HandPose = se.dataclass_field(lambda: se.IntEnum(HandPose, se.U32), default=0)
    joints: OrderedMultiDict[str, Joint] = se.dataclass_field(se.MultiDictAdapter(
        se.Collection(se.U32, se.Tuple(se.CStr(), JOINT_DATACLASS)),
    ))
    constraints: List[Constraint] = se.dataclass_field(
        se.Collection(se.S32, CONSTRAINT_DATACLASS),
    )

    def to_bytes(self):
        writer = se.BufferWriter("<")
        writer.write(se.Dataclass(type(self)), self)
        return writer.copy_buffer()

    @classmethod
    def from_bytes(cls, buff) -> "Animation":
        reader = se.BufferReader("<", buff)
        return reader.read(se.Dataclass(cls))


@dataclasses.dataclass
class Joint:
    priority: int = se.dataclass_field(se.S32)
    rot_keyframes: List[RotKeyframe] = se.dataclass_field(
        se.Collection(se.S32, ROTKEYFRAME_DATACLASS)
    )
    pos_keyframes: List[PosKeyframe] = se.dataclass_field(
        se.Collection(se.S32, POSKEYFRAME_DATACLASS),
    )


def _get_version_from_context(ctx: se.ParseContext) -> Tuple[int, int]:
    return ctx._root.major_version, ctx._root.minor_version


@dataclasses.dataclass
class RotKeyframe:
    time: float = se.dataclass_field(lambda: VERSIONED_TIME)
    rot: Quaternion = se.dataclass_field(se.ContextSwitch(
        _get_version_from_context,
        {
            (0, 1): se.PackedQuat(se.Vector3),
            (1, 0): se.PackedQuat(se.Vector3U16(-1.0, 1.0)),
        },
    ))


@dataclasses.dataclass
class PosKeyframe:
    time: float = se.dataclass_field(lambda: VERSIONED_TIME)
    pos: Vector3 = se.dataclass_field(se.ContextSwitch(
        _get_version_from_context,
        {
            (0, 1): se.Vector3,
            (1, 0): se.Vector3U16(-5.0, 5.0),
        },
    ))


class QuantizedTime(se.QuantizedFloatBase):
    def __init__(self, prim_spec: se.SerializablePrimitive):
        super().__init__(prim_spec, zero_median=False)

    def _get_upper_limit(self, ctx: se.ParseContext) -> float:
        # Upper limit is the "duration" from the parent animation
        return ctx._root.duration

    def encode(self, val: Any, ctx: Optional[se.ParseContext]) -> Any:
        return self._float_to_quantized(val, 0.0, self._get_upper_limit(ctx))

    def decode(self, val: Any, ctx: Optional[se.ParseContext], pod: bool = False) -> Any:
        return self._quantized_to_float(val, 0.0, self._get_upper_limit(ctx))


VERSIONED_TIME = se.ContextSwitch(
    _get_version_from_context,
    {
        (0, 1): se.F32,
        (1, 0): QuantizedTime(se.U16),
    },
)


class ConstraintType(enum.IntEnum):
    POINT = 0
    PLANE = 1


@dataclasses.dataclass
class Constraint:
    chain_length: int = se.dataclass_field(se.U8)
    type: ConstraintType = se.dataclass_field(se.IntEnum(ConstraintType, se.U8))
    source_volume: str = se.dataclass_field(se.StrFixed(16))
    source_offset: Vector3 = se.dataclass_field(se.Vector3)
    target_volume: str = se.dataclass_field(se.StrFixed(16))
    target_offset: Vector3 = se.dataclass_field(se.Vector3)
    target_dir: Vector3 = se.dataclass_field(se.Vector3())
    ease_in_start: float = se.dataclass_field(se.F32)
    ease_in_stop: float = se.dataclass_field(se.F32)
    ease_out_start: float = se.dataclass_field(se.F32)
    ease_out_stop: float = se.dataclass_field(se.F32)


class HandPose(enum.IntEnum):
    SPREAD = 0
    RELAXED = 1
    POINT = 2
    FIST = 3
    RELAXED_L = 4
    POINT_L = 5
    FIST_L = 6
    RELAXED_R = 7
    POINT_R = 8
    FIST_R = 9
    SALUTE_R = 10
    TYPING = 11
    PEACE_R = 12
    PALM_R = 13
